﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.IO;
using System.Threading;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles; 

using EcatLib;
using EcatDef;
using EcatError;

namespace ECAT_M801_DEMO_IO
{
    public partial class Form1 : Form
    {
        [DllImport("kernel32.dll",
            EntryPoint = "GetStdHandle",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        private static extern IntPtr GetStdHandle(int nStdHandle);
        [DllImport("kernel32.dll",
            EntryPoint = "AllocConsole",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        private static extern int AllocConsole();
        private const int STD_OUTPUT_HANDLE = -11;
        private const int MY_CODE_PAGE = 437;

        // copy byte array to struct
        //T ByteArrayToStructure<T>(byte[] bytes) where T : struct
        //{
        //    T stuff;
        //    GCHandle handle = GCHandle.Alloc(bytes, GCHandleType.Pinned);
        //    try
        //    {
        //        stuff = (T)Marshal.PtrToStructure(handle.AddrOfPinnedObject(), typeof(T));
        //    }
        //    finally
        //    {
        //        handle.Free();
        //    }
        //    return stuff;
        //}

        public Form1()
        {
            InitializeComponent();

            InitConsole();

            lstV_SlaveList.View = View.Details;
            lstV_SlaveList.Columns.Add("NO.");
            lstV_SlaveList.Columns.Add("Slave Name.");
            lstV_SlaveList.Columns[0].Width = (int)((double)lstV_SlaveList.Width * 0.2);
            lstV_SlaveList.Columns[1].Width = (int)((double)lstV_SlaveList.Width * 0.8);            
            lstV_SlaveList.BeginUpdate();

            Variables.isOP = false;
            int ret;

            while (true)
            {
                ret = InitDevice();
                if (ret < 0) break;
                ret = Opendevice();
                if (ret < 0) break;
                ret = Startoptask();
                if (ret < 0) break;
                ret = InitModules();
                if (ret < 0) break;
                break;
            }
        }

        private void InitConsole()
        {
            AllocConsole();
            IntPtr stdHandle = GetStdHandle(STD_OUTPUT_HANDLE);
            SafeFileHandle safeFileHandle = new SafeFileHandle(stdHandle, true);
            FileStream fileStream = new FileStream(safeFileHandle, FileAccess.Write);
            Encoding encoding = System.Text.Encoding.GetEncoding(MY_CODE_PAGE);
            StreamWriter standardOutput = new StreamWriter(fileStream, encoding);
            standardOutput.AutoFlush = true;
            Console.SetOut(standardOutput);
            Console.WriteLine("This text you can see in console window.");
        }

        public bool timeoutfunc(DateTime startTime)
        {
            DateTime currTime = DateTime.Now;
            TimeSpan diff_st = new TimeSpan(currTime.Ticks - startTime.Ticks);

            if (diff_st.TotalMilliseconds > Constants.TIME_OUT)
            {               
                return true;
            }
            else return false;
        }

        public int InitDevice()
        {
	        int ret;
	        ushort DeviceCnt = 0;
            byte[] CardId = new byte[CEcatDef.CARD_DEVICE_NO_MAX];
            for (int i = 0; i < CardId.Count(); i++)
            {
                CardId[i] = 0;
            }
            ret = CEcatLib.ECAT_GetDeviceCnt(ref DeviceCnt, CardId);

	      	if(ret < 0)
	        {
		        text_state.Text = "InitDevice Failed";
		        return -1;
	        }
	        else
	        {
                if (DeviceCnt == 0)
                {
                    text_state.Text = "InitDevice Failed";
                    return -1;
                }
                text_state.Text = "InitDevice";
		        Variables.DeviceNo=CardId[0];	//Get card id of first ECAT-M801 //there is a switch on ECAT-M801 can change card id
		        return 0;
	        }
        }

        private int Opendevice()
        {
	        int ret = 0;
            string str1;
	      
	        ushort Alias = 0;
	        uint ProductCode = 0, VendorID = 0, RevisionNo = 0, SerialNo = 0, SlaveType = 0;
	        byte  AlState = 0;

            ret = CEcatLib.ECAT_OpenDevice(Variables.DeviceNo);//Open first ECAT-M801 via card id

	        if(ret < 0)
	        {
                str1 = "Open Device failed, DeviceNo:" + Variables.DeviceNo.ToString() + ", ret:" + ret.ToString();		        
                text_state.Text = str1;		        
		        return -1;
	        }
	        else
	        {
                uint LinkUp = 0, SlavesResp = 0, AlStates = 0, Wc = 0;
                CEcatLib.ECAT_StopDeviceOpTask(Variables.DeviceNo);
                for (ushort i = 0; i < 10; i++)
                {
                    ret = CEcatLib.ECAT_GetDeviceStateEx(Variables.DeviceNo, ref LinkUp, ref SlavesResp, ref AlStates, ref Wc);
                    if (ret == 0 && AlStates == CEcatDef.ECAT_AS_PREOP)
                        break;
                    Thread.Sleep(1000);
                }
                if (ret != 0 || AlStates != CEcatDef.ECAT_AS_PREOP)
                {
                    str1 = "ECAT_GetDeviceStateEx failed, State:" + SlavesResp.ToString() + ", ret:" + ret.ToString();
                    text_state.Text = str1;
                    return -1;
                }
                text_state.Text = "Open Device";
                timer1.Interval = Constants.DEV_TIMER_CYCLE_TIME;
                timer1.Enabled = true;

                ret = CEcatLib.ECAT_SetSlaveNoType(Variables.DeviceNo, CEcatDef.SLAVE_NO_TYPE_POSITION);
                if (ret != 0)
                {
                    str1 = "SetSlaveNoType failed, DeviceNo:" + Variables.DeviceNo.ToString() + ", ret:" + ret.ToString();
                    text_state.Text = str1;
                    return -1;
                }

                StringBuilder slvName = new StringBuilder("", CEcatDef.MAX_SLAVE_NAME_LENGTH);

		        //Get status of ECAT-M801 
                ret = CEcatLib.ECAT_GetDeviceStateEx(Variables.DeviceNo, ref LinkUp, ref SlavesResp, ref AlStates, ref Wc);
		        
		        if(ret == 0)
		        {
                    lstV_SlaveList.Items.Clear();

			        for(ushort i=0;i<SlavesResp;i++)
			        {				                                                        
				        //Get Slave information via the position of the module in the ethercat network
                        ret = CEcatLib.ECAT_GetSlaveInfo(Variables.DeviceNo, i, ref Alias, ref ProductCode,
                                ref VendorID, ref RevisionNo, ref SerialNo, ref AlState, ref SlaveType, slvName);
				    
                        lstV_SlaveList.Items.Add(i.ToString("D"), i.ToString("D"), 0); 
                        if (ret == 0)
                            lstV_SlaveList.Items[i.ToString("D")].SubItems.Add(slvName.ToString());				      
			        }
		        }
                ret = CEcatLib.ECAT_SetSlaveNoType(Variables.DeviceNo, CEcatDef.SLAVE_NO_TYPE_ALIAS_ID);
                if (ret != 0)
                {
                    str1 = "SetSlaveNoType failed, DeviceNo:" + Variables.DeviceNo.ToString() + ", ret:" + ret.ToString();
                    text_state.Text = str1;
                    return -1;
                }
                return 0;
	        }
        }

        private int InitModules()
        {
            int ret = 0;
            string str1 = "";

            ushort Index;		//Home Offset Index
            byte SubIndex;	//Home Offset SubIndex
            ushort DataSize;	//DataSize
            uint ObjectVal;	//Value
            uint AbortCode = 0; //AbortCode
            ushort ChannelNo = 0;//ChannelNo

            //ECAT-2011H ChannelNo 0 set Differential 
            Index = (ushort)(0x8000 + 0x10 * ChannelNo);		//AI Setting
            SubIndex = 0x09;	//Analog input type
            DataSize = 1;
            ObjectVal = 0;		//Differential

            ret = CEcatLib.ECAT_SetSlaveSdoObject(Variables.DeviceNo, Constants.ecat_2011h_SlaveNo, Index, SubIndex,
                DataSize, ObjectVal, ref AbortCode);
            if (ret != 0)
            {
                str1 = "Failed to Set ECAT-2011H, ret:" + ret.ToString() + ", AbortCode:" + AbortCode.ToString();
                Console.WriteLine(str1);                                
                return -1;
            }

            //ECAT-2011H ChannelNo 0 set AI range
            byte Range = CEcatDef.SLAVE_AI_BI_10V;//+-10V
            ret = CEcatLib.ECAT_SetSlaveAiProperty(Variables.DeviceNo, Constants.ecat_2011h_SlaveNo, ChannelNo, Range);
            if (ret != 0)
            {
                str1 = "Failed to Set ECAT-2028, ret:" + ret.ToString();
                Console.WriteLine(str1);
                return -1;
            }

            //ECAT-2028 set AO range
            Range = CEcatDef.SLAVE_AO_BI_10V;//+-10V
            ret = CEcatLib.ECAT_SetSlaveAoProperty(Variables.DeviceNo, Constants.ecat_2028_SlaveNo, ChannelNo, Range);
            if (ret != 0)
            {
                str1 = "Failed to Set ECAT-2028, ret:" + ret.ToString();
                Console.WriteLine(str1);
                return -1;
            }
            return 0;
        }

        private int Startoptask()
        {
            int ret = 0;            
            ushort NetworkInfoNo = 0;
            byte EnumCycleTime = 0;//1ms
            uint WcErrCnt=3;//Tolerance of Ethercat communication Error 

            for (ushort i = 0; i < 10; i++)
            {
                ret = CEcatLib.ECAT_StartDeviceOpTask(Variables.DeviceNo, NetworkInfoNo, EnumCycleTime, WcErrCnt);
                if (ret == 0)
                    break;
                Thread.Sleep(1000);
            }

            if (ret < 0)
            {
                if (ret == -1009)
                    text_state.Text = "Op Failed, please use Utility.exe to build Network Information";
                else
                    text_state.Text = "Op Failed";
                return -1;
            }
            else
            {
                uint LinkUp = 0, SlavesResp = 0, AlStates = 0, Wc = 0;
                text_state.Text = "Oping...";
                for (ushort i = 0; i < 30; i++)
                {
                    ret = CEcatLib.ECAT_GetDeviceStateEx(Variables.DeviceNo, ref LinkUp, ref SlavesResp, ref AlStates, ref Wc);
                    if (ret == 0 && AlStates == CEcatDef.ECAT_AS_OP)
                        break;
                    Thread.Sleep(1000);
                }
                if (ret != 0 || AlStates != CEcatDef.ECAT_AS_OP)
                {
                    text_state.Text = "Op Failed";
                    return -1;
                }
                else
                {
                    text_state.Text = "OP";
                    return 0;
                }
            }
        }

        private void ModuleProcess()
        {
	        int ret;
            byte[] bData = new byte[Constants.rcv_ByteSize];
            byte[] bRcvData = new byte[Constants.rcv_ByteSize];
            ushort ChannelNo = 0;
            double dValue = 0;
            uint U32Value = 0;
            //ECAT_2011H //V0 //Channel 0
            //get voltage
            ret = CEcatLib.ECAT_GetSlaveAiVoltData(Variables.DeviceNo, Constants.ecat_2011h_SlaveNo, ChannelNo, ref dValue);
            if (ret == 0)
            {
                text_E2011H_IN.Text = dValue.ToString("#0.000");
            }

            //get current
            //ret = CEcatLib.ECAT_GetSlaveAimAData(Variables.DeviceNo, Constants.ecat_2011h_SlaveNo, ChannelNo, ref dValue);
            //if (ret == 0)
            //{
            //    text_E2011H_IN.Text = dValue.ToString("#0.000");
            //}

            //ECAT_2028 //VOUT0
            ret = CEcatLib.ECAT_GetSlaveAoVoltData(Variables.DeviceNo, Constants.ecat_2028_SlaveNo, ChannelNo, ref dValue);
	        if(ret == 0)
	        {
                text_E2028_OUT.Text = dValue.ToString("#0.000");
	        }
        
	        //ECAT_2051-32
            ret = CEcatLib.ECAT_GetSlaveDI(Variables.DeviceNo, Constants.ecat_2051_SlaveNo, ref U32Value);
	        if(ret == 0)
	        {
                text_E2051_IN.Text = U32Value.ToString();                
	        }

	        //ECAT_2045-32            
	        ret = CEcatLib.ECAT_GetSlaveDO(Variables.DeviceNo, Constants.ecat_2045_SlaveNo, ref U32Value);
	        if(ret == 0)
	        {
                text_E2045_OUT.Text = U32Value.ToString();                
	        }	        
        }        

        private void UpdateDeviceState(uint LinkUp, uint SlavesResp, uint AlStates, uint Wc)
        {
		    string strAlState;
                text_SlavesResp.Text = "Slave(s):" + SlavesResp.ToString();

			    if(AlStates == 0)
			    {
				    strAlState = "INIT"; 
				    Variables.isOP=false;
			    }
			    else if(AlStates == 0x02)
			    {
				    strAlState = "PREOP";  
				    Variables.isOP=false;
			    }
			    else if(AlStates == 0x04)
			    {
				    strAlState = "SAFEOP";
				    Variables.isOP=false;
			    }
			    else if(AlStates == 0x08)
			    {
				    if(Variables.isOP == false)
				    {
                        text_AlStates.Text = "Op";					    
				    }
				    strAlState = "OP";
				    Variables.isOP=true;
			    }
			    else
			    {
				    strAlState = "Unknown";
				    Variables.isOP=false;
			    }

                text_AlStates.Text = strAlState;
                text_Wc.Text = "WC:" + Wc.ToString();			    			    

			    if(LinkUp != 0)
			    {
                    text_LinkUp.Text = "Up";				    
			    }
			    else
			    {
                    text_LinkUp.Text = "Down";				    
			    }
		    }
        private void timer1_Tick(object sender, EventArgs e)
        {
            int ret;
		    string strAlState;
		    uint LinkUp = 0, SlavesResp = 0, AlStates = 0, Wc = 0;

		    ret = CEcatLib.ECAT_GetDeviceStateEx(Variables.DeviceNo, ref LinkUp, ref SlavesResp, ref AlStates, ref Wc);
		    if(ret == 0)
		    {
                UpdateDeviceState(LinkUp, SlavesResp, AlStates, Wc);
            }
            if (Variables.isOP)
		    {
			    ModuleProcess();
		    }
        }

        private void btn_E2028_SET_Click(object sender, EventArgs e)
        {
            string str1;
            int ret = 0;
            double dValue = Convert.ToDouble(text_E2028_SET.Text);
            ushort ChannelNo = 0;
            
            //ECAT-2028 VOUT0
            ret = CEcatLib.ECAT_SetSlaveAoVoltData(Variables.DeviceNo, Constants.ecat_2028_SlaveNo, ChannelNo, dValue);
            if (ret != 0)
            {
                str1 = "Failed to set ECAT-2028 ,ret:" + ret.ToString();
                Console.WriteLine(str1);
            }
        }

        private void btn_E2045_SET_Click(object sender, EventArgs e)
        {
            string str1;
            int ret = 0;

            uint ecat_2045_32_DO = Convert.ToUInt32(text_E2045_SET.Text);
                                    
            ret = CEcatLib.ECAT_SetSlaveDO(Variables.DeviceNo, Constants.ecat_2045_SlaveNo, ecat_2045_32_DO);

            if (ret != 0)
            {
                str1 = "Failed to set ECAT-2045 ,ret:" + ret.ToString();
                Console.WriteLine(str1);
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            CEcatLib.ECAT_StopDeviceOpTask(Variables.DeviceNo);
            CEcatLib.ECAT_CloseDevice(Variables.DeviceNo);
        }
    }
    
    static class Constants
    {        
        public const int DEV_RM_IS_CLOSE = 0;
        public const int DEV_RM_IS_OPEN = 1;
        public const int DEV_RM_IS_OP_TASK = 2;    

        public const int TIME_OUT = 3;
        public const int DEV_TIMER_CYCLE_TIME = 30; //ms

        public const ushort ecat_2011h_SlaveNo = 0; // The position of the module in the ethercat network
        public const ushort ecat_2028_SlaveNo = 1;
        public const ushort ecat_2051_SlaveNo = 2;
        public const ushort ecat_2045_SlaveNo = 3;

        public const ushort ecat_2011h_DataSize = 4;
        public const ushort ecat_2028_DataSize = 4;
        public const ushort ecat_2051_DataSize = 4;
        public const ushort ecat_2045_DataSize = 4;
        public const ushort rcv_ByteSize = 4; // max of 4 slave's datasize                
    }
    public class Variables
    {
        public static ushort DeviceNo;
        public static bool isOP;
        public static bool bDevOpen;
        public static bool bStateOP;
        public static ushort SlaveCnt;
        public static DateTime startTime;

        public static bool bRecordStart;
        public static uint nRcdCnt;

        public static byte[] DoData;
        public static byte[] DiData;
        public static ushort DoBitSize, DoByteSize;
        public static ushort DiBitSize, DiByteSize;

        public static bool[] rxbitstate;
        public static bool[] enablerx;
        public static bool[] txbitstate;
        public static bool[] enabletx;

        public static int DevRunMode;

        public static double ecat_2011h_Scale; // Scale RawData to Volt/mA
        public static double ecat_2028_Scale; // Scale RawData to Volt

        public static ushort ecat_2011h_Value = 0;
        public static ushort ecat_2028_Output = 0;
        public static uint ecat_2051_32_DI = 0;
        public static uint ecat_2045_32_DO = 0;

        //public struct _ObjectInfo_
        //{
        //    public ushort ByteOffset;
        //    public ushort BitOffset;
        //    public ushort BitSize;
        //    public ushort ByteMask;
        //}

        //public struct _ecat_2011h_st_
        //{
        //    public _ObjectInfo_ Underrange;
        //    public _ObjectInfo_ Overrange;
        //    public _ObjectInfo_ Limit1;
        //    public _ObjectInfo_ Limit2;
        //    public _ObjectInfo_ Error;
        //    public _ObjectInfo_ Gap1;
        //    public _ObjectInfo_ Gap2;
        //    public _ObjectInfo_ TxPDO_State;
        //    public _ObjectInfo_ TxPDO_Toggle;
        //    public _ObjectInfo_ Value;
        //}
        //public static _ecat_2011h_st_ ecat_2011h_t;

        //public struct _ecat_2028_st_
        //{
        //    public _ObjectInfo_ Output;
        //    public _ObjectInfo_ Gap;
        //}
        //public static _ecat_2028_st_ ecat_2028_t;

        //public struct _ecat_2051_32_st_
        //{
        //    public _ObjectInfo_ DI;
        //}
        //public static _ecat_2051_32_st_ ecat_2051_32_t;

        //public struct _ecat_2045_32_st_
        //{
        //    public _ObjectInfo_ DO;
        //}
        //public static _ecat_2045_32_st_ ecat_2045_32_t;

        //public struct ecat_event
        //{
        //    public ushort DeviceNo;
        //    public ushort EventID;
        //}
        //public static ecat_event EcatEvent;
    }
}
